"use strict";
this.name        = "escortdeck_escort";
this.author      = "Norby";
this.copyright   = "2014 Norbert Nagy";
this.licence     = "CC BY-NC-SA 4.0";
this.description = "AI support functions for escorts";


this.$debug = true;
this._landing = function() {this._landing2(this.ship);}

this._landing2 = function(ship) {
    var w = worldScripts.escortdeck;
    var p = player.ship;
    if( !p || !p.isValid ) return;
    ship.target = p;
    var dpos = p.position.add(p.velocity.multiply(2)); //target ahead the player when move
    ship.savedCoordinates = dpos; // must use setDestinationToTarget also
    ship.desiredRange = 0;
    var cd = p.collisionRadius + ship.collisionRadius + 50; //contact distance + a few for the deck
    var d = ship.position.distanceTo(p.position);
    if( d < w.$EscortDeckLandingDist ) { //enough near?
        var landnow = true;
        if( p.addCollisionException ) { //from Oolite v1.81
            p.addCollisionException(ship);
            if( d > cd ) {
                landnow = false; //not so near
                //apply "tractor beam" to pull this fighter to the deck
                var v = p.position.add(p.heading.multiply(-p.collisionRadius)).subtract(ship.position);
                //ship.velocity = v.direction().multiply(200).add(v.multiply(0.3)).add(p.velocity);
                ship.velocity = v.direction().multiply(75).add(v.multiply(0.3)).add(p.velocity);
            }
        }
        if(landnow) {
            var pad = w.$EscortDeckShip.indexOf(ship);
            if( pad > -1 && !w.$EscortDeck_PutShip(false, ship, pad, p, w, true, true) )
                ship.AIState = "DECK"; //landed
            else ship.AIState = "FOLLOW"; //no room or replaced with another ship
//            ship.reportAIMessages = false;
            return;
        }
    }
    //do not attack hostiles during landing
    if( d <= cd ) ship.reactToAIMessage("PLAYER_NEXT");
    else if( d < w.$EscortDeckLandingDist ) ship.reactToAIMessage("PLAYER_NEAR");
    else {
        if (ship.fuel >= 0.5) {
            if( ship.injectorSpeedFactor > 0 )
                ship.desiredSpeed = ship.injectorSpeedFactor * ship.maxSpeed;
            else ship.desiredSpeed = 7 * ship.maxSpeed; //for Oolite 1.80
        } else ship.desiredSpeed = ship.maxSpeed;
        ship.reactToAIMessage("PLAYER_FAR");
    }
//    ship.performFlyToRangeFromDestination();
}

this._locatePlayer = function() {this._locatePlayer2(this.ship);}

this._locatePlayer2 = function(ship) {
    var p = player.ship;
    if( !p || !p.isValid ) return;
    var sc = ship.script;
    ship.target = p;
    ship.savedCoordinates = p.position;
    var d = ship.position.distanceTo(p.position);
    if (player.alertCondition === 3 && sc._inEscortFleet2(ship)  && d < 25600) {
        //if not in fleet then follow but not defend the player, just himself
        if (this.$debug) log("escortdeck_escortdeck", "_locatePlayer: "+ship.displayName+" target:"+ p.displayName+" distance:"+d.toFixed(0)+", GOTO");
        ship.AIState = "GOTO"; //attack hostiles
    } else {
        if( d < 1500 ) {
            if (this.$debug) log("escortdeck_escortdeck", "_locatePlayer: "+ship.displayName+" target:"+ p.displayName+" distance:"+d.toFixed(0)+", PLAYER_NEXT_1");
            ship.reactToAIMessage("PLAYER_NEXT_1");
        } else if ( d < 5000 ) {
            if (this.$debug) log("escortdeck_escortdeck", "_locatePlayer: "+ship.displayName+" target:"+ p.displayName+" distance:"+d.toFixed(0)+", PLAYER_NEAR_1_5");
            ship.reactToAIMessage("PLAYER_NEAR_1_5");
        } else if (d <= 15000) {
            if (this.$debug) log("escortdeck_escortdeck", "_locatePlayer: "+ship.displayName+" target:"+ p.displayName+" distance:"+d.toFixed(0)+", PLAYER_FAR_5_15");
            ship.reactToAIMessage("PLAYER_FAR_5_15");
        } else if (d <= 25000) {
            if (this.$debug) log("escortdeck_escortdeck", "_locatePlayer: "+ship.displayName+" target:"+ p.displayName+" distance:"+d.toFixed(0)+", PLAYER_FAR_15_25");
            ship.reactToAIMessage("PLAYER_FAR_15_25");
        } else {
            if (this.$debug) log("escortdeck_escortdeck", "_locatePlayer: "+ship.displayName+" target:"+ p.displayName+" distance:"+d.toFixed(0)+", PLAYER_FAR_25_");
            ship.reactToAIMessage("PLAYER_FAR_25_");
        }
    }
}

this._combatCheck = function() {this._combatCheck2(this.ship);}

this._combatCheck2 = function(ship) {
    var p = player.ship;
    if( !p || !p.isValid ) return;
    var sc = ship.script;

    if(ship.target && ship.target.isDerelict)
    {
        ship.target = null;
        ship.reactToAIMessage("TARGET_EJECTED");
    }
    
    var t = sc._playerTarget2(ship);
    if( !ship.target || !ship.target.isValid
        || t && t.isValid && ship.target != t ) {
        ship.AIState = "GOTO"; //find new target or player set new target
        return;
    }
    
    if( ship.position.distanceTo(p.position) > 24000 ) {
        // if escort is far from the player, break off combat
        if( ship.target && ship.target != sc._findPlayerNearestHostile2(ship) ) {
            ship.target = null;
            ship.AIState = "GOTO"; //change target to the player's nearest hostile
        }
    } else {
        var e = ship.energy;
        var m = ship.maxEnergy;
        ship.savedCoordinates = ship.target.position;
        if( e < m * 0.4 ) //cautious: flee if energy is below 40%
            ship.AIState="FLEE"; //avoid firing ECM
        else if( sc._Inject2(ship.target,ship) ) { //inject?
            if( ship.AIState == "INJECT" )
                ship.reactToAIMessage("ENEMY_FIRE");
            else ship.AIState = "INJECT"; //switch to inject
        } else {
            if( ship.AIState == "ATTACK" )
                ship.reactToAIMessage("ENEMY_FIRE");
            else ship.AIState = "ATTACK"; //no more inject
        }
    }
}

this._findPlayerHostiles = function(caller) {
    if (this.$debug) log(this.name, this.ship.displayName+" _findPlayerHostiles called from "+caller);
    this._findPlayerHostiles2(this.ship);
}

this._findPlayerHostiles2 = function(ship) {
    var p = player.ship;
    if( !p || !p.isValid ) return;
    var sc = ship.script;

    if( !sc._inEscortFleet2(ship) ) { //follow but not defend the player, just himself
        ship.target = null;
        if (this.$debug) log("escortdeck_escortdeck", ship.displayName+" - findPlayerHostiles: not acting as escort");
        ship.reactToAIMessage("NO_HOSTILE_FOUND");
    }
    if (this.$debug) log("escortdeck_escortdeck", ship.displayName+" - findPlayerHostiles: player's target: "+(p.target ? p.target.displayName+" at "+ship.position.distanceTo(p.target).toFixed(1) : "None"));

    if (ship.target && ship.target.isValid && sc._isHostileToPlayer(ship.target)) {
        if (this.$debug) log("escortdeck_escortdeck", ship.displayName+" - findPlayerHostiles: keeping the current target: "+ship.target.displayName+", at "+ship.position.distanceTo(ship.target).toFixed(1));
        sc._foundPlayerHostile2( ship.target, ship ); //keep the current target
        return;
    }

    var hostile = sc._findPlayerNearestHostile2(ship);
    if (hostile) {
        if (this.$debug) log("escortdeck_escortdeck", ship.displayName+" - findPlayerHostiles: targeting hostile closest to player: "+hostile.displayName+" at "+ship.position.distanceTo(hostile));
        sc._foundPlayerHostile2(hostile, ship);
    } else {
        var t = sc._playerTarget2(ship); //follow target changes of player
        if (this.$debug) log("escortdeck_escortdeck", ship.displayName+" - findPlayerHostiles: :_playerTarget2 returned "+t);
        if (t && t.isValid) {
            if (this.$debug) log("escortdeck_escortdeck", ship.displayName+" - findPlayerHostiles: targeting player's target: "+t.displayName);
            sc._foundPlayerHostile2(t, ship);
            return;
        }

        if (this.$debug) log("escortdeck_escortdeck", ship.displayName+" - findPlayerHostiles: No hostiles found");
        ship.target = p;
        ship.reactToAIMessage("NO_HOSTILE_FOUND");
    }
}

this._findPlayerNearestHostile = function() {this._findPlayerNearestHostile2(this.ship);}

this._findPlayerNearestHostile2 = function(ship) {
    var sc = ship.script;
    var pship = player.ship;  
    var hostile = null;
    var distance, distance_tmp;
    var targets = pship.checkScanner(true);//much faster than filteredEntities
    if (targets && targets.length > 0) {
        var i = targets.length;
        while (i--) {
            var t = targets[i];
            if( t && t.isValid && sc._isHostileToPlayer( t ) ) {
                distance_tmp = t.position.distanceTo(pship);
                if (!distance || (distance && distance > distance_tmp)) {
                    distance = distance_tmp;
                    hostile = t;
                }
            }
        }
    }
    return(hostile);
}

this._fireCheck = function() {this._fireCheck2(this.ship);}

this._fireCheck2 = function(ship) {
    var t = ship.target;
    if( !t || !t.isValid || t.isPlayer || (t.script && ( t.script.$EscortDeckUsable > 0
        || t.script.$CarriersShipStorageID > -1 ) ) ) {
        if (this.$debug) log(this.name, ship.displayName+" _fireCheck, FRIENDLY_FIRE, target: "+t.displayName+", escort:"+t.script.$EscortDeckUsable+", carrierEscort:"+t.script.$CarriersShipStorageID);
        ship.target = null;
        ship.reactToAIMessage("FRIENDLY_FIRE");
    } else {
        if (this.$debug) log(this.name, ship.displayName, " energy "+ship.energy+", _fireCheck, ENEMY_FIRE, target: "+t.displayName+" at "+ship.position.distanceTo(t));
        ship.reactToAIMessage("ENEMY_FIRE");
    }
}

this._fleeCheck = function() {this._fleeCheck2(this.ship);}

this._fleeCheck2 = function(ship) {
    var e = ship.energy;
    var m = ship.maxEnergy;
    if( (!ship.target || !ship.target.isMissile) && e > m * 0.8 ) {
        //not a missile run and recovered 80% energy
        ship.AIState = "GOTO";
    }
}

this._foundPlayerHostile = function( hostile ) {this._foundPlayerHostile2(hostile, this.ship);}

this._foundPlayerHostile2 = function( hostile, ship ) {
    var sc = ship.script;

    if( !hostile || !hostile.isValid) {
        ship.AIState = "FOLLOW"; //find new target
        return;
    }
    
    ship.target = hostile;
    ship.savedCoordinates = hostile.position;
    var sniper = false;
    if( sc._isSniper2(ship) ) {
        if (this.$debug) log("escortdeck_escortdeck", ship.displayName+" - foundPlayerHostile: "+hostile.displayName+" at "+ship.position.distanceTo(hostile)+", SNIPING");
        ship.reactToAIMessage("SNIPING");
    } else if(sc._Inject2(hostile, ship)) {
        if (this.$debug) log("escortdeck_escortdeck", ship.displayName+" - foundPlayerHostile: "+hostile.displayName+" at "+ship.position.distanceTo(hostile)+", HOSTILE_FAR");
        ship.reactToAIMessage("HOSTILE_FAR"); //inject
    } else {
        if (this.$debug) log("escortdeck_escortdeck", ship.displayName+" - foundPlayerHostile: "+hostile.displayName+" at "+ship.position.distanceTo(hostile)+", HOSTILE_FOUND");
        ship.reactToAIMessage("HOSTILE_FOUND"); //attack
    }
}

this._inEscortFleet = function() {this._inEscortFleet2(this.ship);}

this._inEscortFleet2 = function(ship) { //if not then follow but not defend the player, just himself
    var w = worldScripts.escortdeck;
    if( w.$EscortDeckShip.indexOf(ship) > -1 ) return true;
    return false;
}

this._isHostileToPlayer = function(entity) {
    return (entity.isThargoid || (entity.isShip && !entity.isPlayer && entity.target
        && !entity.isDerelict && ( entity.target == player.ship && entity.hasHostileTarget
            || entity.script && entity.script.$EscortDeck_Target )
        && ( !entity.script || !entity.script.$CarriersShipStorageID //is not owned by player
            || entity.script.$CarriersShipStorageID == -1
            || !( entity.script.$EscortDeckUsable > 0 ))));
}

this._isSniper = function() {this._isSniper2(this.ship);}

this._isSniper2 = function(ship) {
    var l = ship.forwardWeapon.equipmentKey;
    if( l === "EQ_WEAPON_MILITARY_LASER" ) return true;
    else if( l ) {  //Oolite 1.81 can define custom lasers
        var weapon_info = ship.forwardWeapon.weaponInfo;
        if (weapon_info && weapon_info.range > 15000 ) return true;
    }
    return false;
}

this._Inject = function(target) {this._Inject2(target, this.ship);}

this._Inject2 = function(target, ship) { //can use injectors and farther than 10km?
    if(ship && target && target.isValid && ship.fuel > 0
        && ship.equipmentStatus("EQ_FUEL_INJECTION") === "EQUIPMENT_OK"
        && ship.position.distanceTo(target) > 10000 ) {
        //more conditions: not enogh good accuracy, faster target or short range laser?
        if ( ship.accuracy < 7.5 || target.speed > ship.maxspeed )
            return true;
        var l = ship.forwardWeapon.equipmentKey;
        if( l === "EQ_WEAPON_PULSE_LASER" ) return true;
        else if( l ) {  //Oolite 1.81 can define custom lasers
            var weapon_info = ship.forwardWeapon.weaponInfo;
            if (weapon_info && weapon_info.range < 15000 ) return true;
        }
    }
    return false;
}

this._playerTarget = function() {this._playerTarget2(this.ship);}

this._playerTarget2 = function(ship) {
    var t = player.ship.target;
    if( t && t.isValid && !t.isDerelict && ( !t.script || !( t.script.$EscortDeck_Usable > 0 ) )
        && ( t.bounty > 2 || t.script && t.script.$EscortDeck_Target ) && //nonclean or player fired it
        t.scanClass !== "CLASS_CARGO" // avoid escape capsules
//        && ship.position.distanceTo(t.position) < ship.scannerRange )
      )
        return(t);
    return(null);
}


//ship script events
this.shipCollided = function(otherShip) {
    var w = worldScripts.escortdeck;
    if( otherShip == player.ship && this.ship.energy < 50
        && !this.ship.addCollisionException //before Oolite 1.81
        && w.$EscortDeckShip.indexOf(this.ship) > -1 )
        this.ship.energy = 50; //prevent kill owned escorts by accident
}

this.shipTargetDestroyed = function(target) {
    if(this.ship == player.ship) return; //exit if worldScript, do in ship script only
    if(target.primaryRole == "constrictor" && missionVariables.conhunt && missionVariables.conhunt == "STAGE_1") {
        // just in case an escort kills the constrictor, let's not break the mission for the player...
        missionVariables.conhunt = "CONSTRICTOR_DESTROYED";
    }
    
    if(target.isRock || target.isBoulder || target.isCargo || target.isDerelict || target.isWeapon)
        return;
    
//    player.score += 1; - score should remain personal
    player.credits += target.bounty;
    if( this.$EscortDeckUsable > 0 || this.$CarriersShipStorageID > -1 )
        player.consoleMessage(this.ship.displayName+" killed "
            + target.name + " : " + target.bounty + "₢ awarded.", 5);
    log(this.name, this.ship.displayName+" killed " + target.name + " : " + target.bounty +"₢");
}

this.shipDied = function(whom,why) {
    if(this.ship == player.ship) return; //exit if worldScript, do in ship script only
    var w = worldScripts.escortdeck;

    //increase steering and acceleration in proportion with the smaller mass on deck
    w.$EscortDeck_SetMaxs( player.ship, w );

    var pad = w.$EscortDeckShip.indexOf(this.ship);
    if( pad > -1 ) w.$EscortDeckShipData[pad] = null; //prevent recreation in dock
    
    if( whom && whom.isValid && (this.$EscortDeckUsable > 0
        || this.$CarriersShipStorageID > -1 )) {
        player.commsMessage(this.ship.displayName+" terminated.", 10);
    }
    log(this.name, this.ship.displayName+" terminated by "+whom+" "+why);
}
